from sage.misc.all import load, save, cached_function
from sage.matrix.all import matrix, identity_matrix
from sage.modular.all import Gamma0
from sage.rings.all import QQ

################################################################################
### inclusion_matrix_level
################################################################################


## this dictionary maps weights to a list of levels
_stored_inclusion_matrix_level = \
    {  0 : [1, 2, 3, 4, 5, 7, 8, 11, 23],
       2 : [1, 2, 3, 4, 5, 7, 8, 11, 23] }

@cached_function
def load_inclusion_matrix_level( level1, level2, weight, order = 0 ) :
    if order == 0 :
        return load( "data/inclusion_matrix_level/level1_{0}__level2_{1}__weight_{2}.sobj".format( level1, level2, weight ) )

    return load_order_isomorphism_matrix( level2, weight, order ).inverse() \
        * load( "data/inclusion_matrix_level/level1_{0}__level2_{1}__weight_{2}.sobj".format( level1, level2, weight + 12 * order ) ) \
        * load_order_isomorphism_matrix( level1, weight, order )


################################################################################
### inclusion_matrix_order
################################################################################


## this dictionary maps triples (weight, order1, order2) to a list of levels
_stored_inclusion_matrix_order = dict()

@cached_function
def load_inclusion_matrix_order( level, weight, order1, order2 ) :
    return load( "data/inclusion_matrix_order/level_{0}__weight_{1}__order1_{2}__order2_{3}.sobj".format( level, weight, order1, order2 ) )


# ################################################################################
# ### order_isomorphism_matrix
# ################################################################################


## this dictionary maps a pair (weights, order) to a list levels for which
## order isomorphism matrices
_stored_order_isomorphism_matrix = dict()

@cached_function
def load_order_isomorphism_matrix( level, weight, order ) :
    return load( "data/order_isomorphism_matrix/level_{0}__weight_{1}__order_{2}".format( level, weight, order ) )


################################################################################
### projected_cusp_expansion_matrices
################################################################################


## this dictionary maps weights to a list of levels
_stored_projected_cusp_expansion_matrices = \
    { 2  : [1, 2, 3, 4, 5, 7, 8, 11, 23] }

@cached_function
def load_projected_cusp_expansion_matrices( level, weight, order = 0 ) :
    if weight + 12 * order == 0 :
        proj_matrices_pre = dict( (cusp, identity_matrix(QQ, 1))
                                  for cusp in Gamma0(level).cusps() )
    else :
        proj_matrices_pre = load( "data/projected_cusp_expansion_matrices/level_{0}__weight_{1}.sobj".format(level, weight + 12 * order) )

    if order == 0 :
        order_isomorphism_matrix = identity_matrix(QQ, proj_matrices_pre.values()[0].nrows())
    else :
        order_isomorphism_matrix = load_order_isomorphism_matrix( level, weight, order )
    order_isomorphism_matrix_inv = order_isomorphism_matrix.inverse()

    proj_matrices = dict()
    for d in proj_matrices_pre.keys() :
        proj_matrices[d] = \
            order_isomorphism_matrix_inv * proj_matrices_pre[d] \
            * order_isomorphism_matrix

    equivalent_cusp = lambda c: filter( lambda c2: c2.is_gamma0_equiv(c, level),
                                        proj_matrices_pre.keys() )[0]
    return (lambda cusp: proj_matrices_pre[equivalent_cusp(cusp)])
